library ieee;
use ieee.STD_LOGIC_1164.all;
use ieee.numeric_std.all;

use work.wishbone_pkg.all;

entity vmecore_test is

  port(
    clk_sys_i : in std_logic;
    rst_n_i   : in std_logic;

    slave_i : in  t_wishbone_slave_in;
    slave_o : out t_wishbone_slave_out;

    leds_o : out std_logic_vector(15 downto 0);
    irq_o : out std_logic
    );

end vmecore_test;

architecture rtl of vmecore_test is
  --  Memory map (WB addresses, multiply by 4 to get VME addresses):
  --  0 - 0x1ff: sram (512*4B)
  --  0x1000: leds (4B)
  --  0x1001: last WB transaction (see the code for the format)
  --  0x1002: nbr of WB read accesses (write to clear)
  --  0x1003: nbr of WB write accesses (likewise)
  --  0x1004: nbr of write errors in pattern ram
  --  0x1005: generates bus error.
  --  0x2000: counter (4B). Generate an interrupt when 0 is reached.
  --  0x3000: pattern ram (0x1000 * 4B)
  signal counter : unsigned(31 downto 0);
  signal leds : std_logic_vector(15 downto 0);
  
  signal last_trans : std_logic_vector (20 downto 0);

  signal nbr_read : unsigned (15 downto 0);
  signal nbr_write: unsigned (15 downto 0);
  signal nbr_write_errors: unsigned (31 downto 0);

  signal pattern : std_logic_vector (31 downto 0);
  
  type sram_type is array (0 to 16#1ff#) of std_logic_vector(31 downto 0);
  signal sram: sram_type;
begin

  pattern (31 downto 16) <= not slave_i.adr(15 downto 0);
  pattern (15 downto 0)  <= slave_i.adr(15 downto 0);

  process (clk_sys_i)
    variable idx : natural;
    variable err : boolean;
  begin
    if rising_edge(clk_sys_i) then
      slave_o.ack <= '0';
      slave_o.stall <= '0';
      slave_o.err <= '0';
      
      if rst_n_i = '0' then
        counter <= (others => '0');
        leds <= (others => '0');
        nbr_read <= (others => '0');
        nbr_write <= (others => '0');
        nbr_write_errors <= (others => '0');
      else
        -- Decrementer
        if counter /= (counter'range => '0') then
          counter <= counter - 1;
        end if;

        if slave_i.stb = '1' then
          if slave_i.adr (13 downto 12) = "00" then
            last_trans (15 downto 0) <= slave_i.adr (15 downto 0);
            last_trans (19 downto 16) <= slave_i.sel;
            last_trans (20) <= slave_i.we;
          end if;

          if slave_i.we = '1' then
            --  Write
            nbr_write <= nbr_write + 1;
            case slave_i.adr (13 downto 12) is
              when "00" =>
                idx := to_integer(unsigned(slave_i.adr(8 downto 0)));
                for i in 3 downto 0 loop
                  if slave_i.sel (i) = '1' then
                    sram(idx)(8*i + 7 downto 8*i) <=
                      slave_i.dat(8*i + 7 downto 8*i);
                  end if;
                end loop;
              when "01" =>
                case slave_i.adr (2 downto 0) is
                  when "000" =>
                    for i in 1 downto 0 loop
                      if slave_i.sel (i) = '1' then
                        leds(8*i + 7 downto 8*i) <=
                          slave_i.dat(8*i + 7 downto 8*i);
                      end if;
                    end loop;
                  when "001" =>
                    null;
                  when "010" =>
                    nbr_read <= (others => '0');
                  when "011" =>
                    nbr_write <= (others => '0');
                  when "100" =>
                    nbr_write_errors <= (others => '0');
                  when "101" =>
                    slave_o.err <= '1';
                  when others =>
                    null;
                end case;
              when "10" =>
                for i in 3 downto 0 loop
                  if slave_i.sel (i) = '1' then
                    counter(8*i + 7 downto 8*i) <=
                      unsigned(slave_i.dat(8*i + 7 downto 8*i));
                  end if;
                end loop;
              when "11" =>
                err := false;
                for i in 3 downto 0 loop
                  if slave_i.sel (i) = '1'
                    and (slave_i.dat(8*i + 7 downto 8*i) /=
                         pattern (8*i + 7 downto 8*i))
                  then
                    err := true;
                  end if;
                end loop;
                if err then
                  nbr_write_errors <= nbr_write_errors + 1;
                end if;
              when others =>
                null;
            end case;
            slave_o.ack <= '1';
          else
            -- Read
            nbr_read <= nbr_read + 1;
            case slave_i.adr (13 downto 12) is
              when "00" =>
                idx := to_integer(unsigned(slave_i.adr(8 downto 0)));
                slave_o.dat <= sram(idx);
              when "01" =>
                case slave_i.adr (2 downto 0) is
                  when "000" =>
                    slave_o.dat(31 downto 16) <= (others => '0');
                    slave_o.dat(15 downto 0) <= leds;
                  when "001" =>
                    slave_o.dat <= (31 downto 21 => '0') & last_trans;
                  when "010" =>
                    slave_o.dat (31 downto 16) <= (others => '0');
                    slave_o.dat (15 downto 0) <= std_logic_vector (nbr_read);
                  when "011" =>
                    slave_o.dat (31 downto 16) <= (others => '0');
                    slave_o.dat (15 downto 0) <= std_logic_vector (nbr_write);
                  when "100" =>
                    slave_o.dat <= std_logic_vector (nbr_write_errors);
                  when "101" =>
                    slave_o.err <= '1';
                  when others =>
                    null;
                end case;
              when "10" =>
                slave_o.dat <= std_logic_vector(counter);
              when "11" =>
                slave_o.dat <= pattern;
              when others =>
                null;
            end case;
            slave_o.ack <= '1';
          end if;
        end if;
      end if;
    end if;
  end process;

  leds_o <= leds;
  irq_o <= '1' when counter = 1 else '0';
end rtl;
